#ifndef ORDER2_LOCAL_DT_HYPERELASTIC_SOLVER_HPP
#define ORDER2_LOCAL_DT_HYPERELASTIC_SOLVER_HPP

#include <mesh/MeshTraits.hpp>

#include <memory>
#include <tuple>
#include <vector>

class DiscreteFunctionVariant;
class IBoundaryConditionDescriptor;
class MeshVariant;
class ItemValueVariant;
class SubItemValuePerItemVariant;
class DiscreteFunctionVariant;

class Order2LocalDtHyperelasticSolverHandler
{
 public:
  enum class SolverType
  {
    Glace,
    Eucclhyd
  };

 private:
  struct IOrder2LocalDtHyperelasticSolver
  {
    virtual std::tuple<std::shared_ptr<const MeshVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const MeshVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>>
  apply(const SolverType& solver_type,
        const size_t& law,
        const double& dt1,
        const size_t& q,
        const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
        const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
        const std::shared_ptr<const DiscreteFunctionVariant>& u1,
        const std::shared_ptr<const DiscreteFunctionVariant>& u2,
        const std::shared_ptr<const DiscreteFunctionVariant>& E1,
        const std::shared_ptr<const DiscreteFunctionVariant>& E2,
        const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
        const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
        const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
        const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
        const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
        const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
        const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
        const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
        const std::shared_ptr<const DiscreteFunctionVariant>& p1,
        const std::shared_ptr<const DiscreteFunctionVariant>& p2,
        const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list1,
        const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list2,
        const std::shared_ptr<const DiscreteFunctionVariant>& lambda1,
        const std::shared_ptr<const DiscreteFunctionVariant>& mu1,
        const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
        const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
        const std::shared_ptr<const DiscreteFunctionVariant>& gamma1,
        const std::shared_ptr<const DiscreteFunctionVariant>& p_inf1,
        const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
        const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2) const = 0;

    virtual std::tuple<std::shared_ptr<const MeshVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const MeshVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>,
                       std::shared_ptr<const DiscreteFunctionVariant>>
  apply(const SolverType& solver_type,
        const size_t& law,
        const double& dt1,
        const std::shared_ptr<const DiscreteFunctionVariant>& rho1,
        const std::shared_ptr<const DiscreteFunctionVariant>& rho2,
        const std::shared_ptr<const DiscreteFunctionVariant>& u1,
        const std::shared_ptr<const DiscreteFunctionVariant>& u2,
        const std::shared_ptr<const DiscreteFunctionVariant>& E1,
        const std::shared_ptr<const DiscreteFunctionVariant>& E2,
        const std::shared_ptr<const DiscreteFunctionVariant>& CG1,
        const std::shared_ptr<const DiscreteFunctionVariant>& CG2,
        const std::shared_ptr<const DiscreteFunctionVariant>& aL1,
        const std::shared_ptr<const DiscreteFunctionVariant>& aL2,
        const std::shared_ptr<const DiscreteFunctionVariant>& aT1,
        const std::shared_ptr<const DiscreteFunctionVariant>& aT2,
        const std::shared_ptr<const DiscreteFunctionVariant>& sigma1,
        const std::shared_ptr<const DiscreteFunctionVariant>& sigma2,
        const std::shared_ptr<const DiscreteFunctionVariant>& p1,
        const std::shared_ptr<const DiscreteFunctionVariant>& p2,
        const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list1,
        const std::vector<std::shared_ptr<const IBoundaryConditionDescriptor>>& bc_descriptor_list2,
        const std::shared_ptr<const DiscreteFunctionVariant>& lambda1,
        const std::shared_ptr<const DiscreteFunctionVariant>& mu1,
        const std::shared_ptr<const DiscreteFunctionVariant>& lambda2,
        const std::shared_ptr<const DiscreteFunctionVariant>& mu2,
        const std::shared_ptr<const DiscreteFunctionVariant>& gamma1,
        const std::shared_ptr<const DiscreteFunctionVariant>& p_inf1,
        const std::shared_ptr<const DiscreteFunctionVariant>& gamma2,
        const std::shared_ptr<const DiscreteFunctionVariant>& p_inf2) const = 0;

    IOrder2LocalDtHyperelasticSolver()                                        = default;
    IOrder2LocalDtHyperelasticSolver(IOrder2LocalDtHyperelasticSolver&&)            = default;
    IOrder2LocalDtHyperelasticSolver& operator=(IOrder2LocalDtHyperelasticSolver&&) = default;

    virtual ~IOrder2LocalDtHyperelasticSolver() = default;
  };

  template <MeshConcept MeshType>
  class Order2LocalDtHyperelasticSolver;

  std::unique_ptr<IOrder2LocalDtHyperelasticSolver> m_hyperelastic_solver;

 public:
  const IOrder2LocalDtHyperelasticSolver&
  solver() const
  {
    return *m_hyperelastic_solver;
  }

  Order2LocalDtHyperelasticSolverHandler(const std::shared_ptr<const MeshVariant>& mesh1,
                                   const std::shared_ptr<const MeshVariant>& mesh2);
};

#endif   // HYPERELASTIC_SOLVER_HPP
